import os
from scikitplot.helpers import cumulative_gain_curve
import matplotlib.pyplot as plt
from sklearn.metrics import f1_score, confusion_matrix, precision_recall_curve, average_precision_score, PrecisionRecallDisplay
import numpy as np
#import mlflow

#def get_mlflow():
#    return mlflow.get_tracking_uri()

def plot_cumulative_gain(y_true, y_proba, title_fontsize=15, text_fontsize=10):
    # Compute Cumulative Gain Curves
    percentages, gains1 = cumulative_gain_curve(y_true, y_proba, True)
    
    # Best classifier
    #percentages, gains2 = cumulative_gain_curve(y_true, y_true, True)

    fig, ax = plt.subplots(1, 1)

    ax.set_title('Cumulative gains chart', fontsize=title_fontsize)

    ax.plot(percentages, gains1, lw=3, label='Class {}'.format(True))
    
    # Best classifier
    #ax.plot(percentages, gains2, lw=3, label='Class {}'.format('best'))

    ax.set_xlim([0.0, 1.0])
    ax.set_ylim([0.0, 1.0])

    ax.plot([0, 1], [0, 1], 'k--', lw=2, label='Baseline')

    ax.set_xlabel('Percentage of sample', fontsize=text_fontsize)
    ax.set_ylabel('Gain', fontsize=text_fontsize)
    ax.tick_params(labelsize=text_fontsize)
    ax.grid('on')
    ax.legend(loc='lower right', fontsize=text_fontsize)
    
    return ax


def get_confusion_matrix(y_true, y_pred_proba, threshold):
    cm = confusion_matrix(y_true, y_pred_proba>threshold)
    return cm

def to_labels(pos_probs, threshold):
    return (pos_probs >= threshold).astype('int')

def get_optimal_f1_cutoff(y_true, y_pred_proba, thresholds = np.arange(0,1, 0.001)):
    scores = [f1_score(y_true, to_labels(y_pred_proba, t)) for t in thresholds]
    ix = np.argmax(scores)
    optimal_threshold = thresholds[ix] 
    optimal_f1_score = scores[ix] 
    
    return optimal_threshold, optimal_f1_score

def get_lift_score(y_true, y_pred_proba, percentile, threshold):
    
    cutoff = round((len(y_pred_proba)*percentile))
    ind = np.argsort(y_pred_proba)[::-1][:cutoff]
    y_pred_sorted = y_pred_proba[ind] > threshold
    y_true_sorted = y_true.values[ind]

    with_model = np.sum(y_true_sorted) / len(y_true_sorted)
    without_model = np.sum(y_true) / len(y_true)
    lift_score = with_model/without_model
    
    return lift_score

def get_threshold_and_cutoff_for_positives(y_true, y_pred_proba, number_of_positives = 1000):
    precisions, recalls, thresholds = precision_recall_curve(y_true, y_pred_proba)

    # find threshold which coincides with the prefered alarm rate
    for threshold in np.flipud(thresholds):
        if (np.sum(y_pred_proba > threshold) >= number_of_positives):
            #print(threshold)
            f = threshold
            break

    cutoff = np.where(thresholds == f)[0][0]
    return threshold, cutoff

def get_partial_ap(y_true, y_pred_proba, number_of_positives=1000):
    precisions, recalls, thresholds = precision_recall_curve(y_true, y_pred_proba)
    _, cutoff = get_threshold_and_cutoff_for_positives(y_true, y_pred_proba, number_of_positives=number_of_positives)
    partial_ap = -np.sum(np.diff(recalls[cutoff:]) * np.array(precisions)[cutoff:-1])
    return partial_ap
    
def plot_partial_ap(y_true, y_pred_proba, number_of_positives=1000):
    precisions, recalls, thresholds = precision_recall_curve(y_true, y_pred_proba)
    _, cutoff = get_threshold_and_cutoff_for_positives(y_true, y_pred_proba, number_of_positives=number_of_positives)

    fig = plt.figure()
    plt.step(recalls[cutoff:], precisions[cutoff:], where='post')

    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.ylim([0.0, 1.05])
    plt.xlim([0.0, 1.0])
    plt.title(
        'Average precision score: AP={0:0.2f}'
        .format(-np.sum(np.diff(recalls[cutoff:]) * np.array(precisions)[cutoff:-1])))

    return fig

def plot_ap(y_true, y_pred_proba):
    # AP curve
    aps = average_precision_score(y_true, y_pred_proba)
    precision, recall, _ = precision_recall_curve(y_true, y_pred_proba)
    disp = PrecisionRecallDisplay(precision=precision, recall=recall, average_precision = aps, estimator_name = None)
    disp.plot()
    return disp.figure_

def get_true_positives_at(y_true, y_pred_proba, number_of_positives = 1000):
    threshold, _ = get_threshold_and_cutoff_for_positives(y_true, y_pred_proba, number_of_positives)
    cm = confusion_matrix(y_true, y_pred_proba>=threshold)
    return cm[1,1]
    
def log_performance(y_true, y_pred_proba, images_path, name,  log_with_mlflow=False, number_of_positives=300):

    performance_dict = {}

    # plot ap curve
    fig = plot_ap(y_true, y_pred_proba)
    plt.savefig(os.path.join(images_path, '_'.join([str(name), 'PRAUCcurve.pdf'])))
    if log_with_mlflow:
        mlflow.log_artifact(os.path.join(images_path, '_'.join([str(name), 'PRAUCcurve.pdf'])))
    
    # Cumulative gains cart
    ax = plot_cumulative_gain(y_true, y_pred_proba)
    fig = plt.gcf()
    plt.savefig(os.path.join(images_path, '_'.join([str(name), 'CumulativeGainsChart.pdf'])))
    if log_with_mlflow:
        mlflow.log_artifact(os.path.join(images_path, '_'.join([str(name), 'CumulativeGainsChart.pdf'])))

    optimal_threshold, optimal_f1_score = get_optimal_f1_cutoff(y_true, y_pred_proba)
    if log_with_mlflow:
        mlflow.log_metric('optimal_threshold', optimal_threshold)
        mlflow.log_metric('f1_score', optimal_f1_score)
    
    performance_dict['optimal_threshold'] = optimal_threshold
    performance_dict['f1_score'] = optimal_f1_score

    # Lift scores
    percentiles = [0.05, 0.01, 0.001]
    for percentile in percentiles:
        lift_score = get_lift_score(y_true, y_pred_proba, percentile, optimal_threshold)
        if log_with_mlflow:
            mlflow.log_metric('lift_score_' + str(percentile), lift_score)
        performance_dict['lift_score_'+ str(percentile)] = lift_score
    
    # confusion matrix
    cm = get_confusion_matrix(y_true, y_pred_proba, optimal_threshold)
    if log_with_mlflow:
        mlflow.log_metric('TN', cm[0,0])
        mlflow.log_metric('FP', cm[0,1])
        mlflow.log_metric('FN', cm[1,0])
        mlflow.log_metric('TP', cm[1,1])    
    
    performance_dict['confusion_matrix/TN'] = cm[0,0]
    performance_dict['confusion_matrix/FP'] = cm[0,1]
    performance_dict['confusion_matrix/FN'] = cm[1,0]
    performance_dict['confusion_matrix/TP'] = cm[1,1]
    

    # partial ap
    partial_ap = get_partial_ap(y_true, y_pred_proba, number_of_positives)
    if log_with_mlflow:
        mlflow.log_metric('partial_ap', partial_ap)
    
    performance_dict['partial_ap'] = partial_ap

    fig = plot_partial_ap(y_true, y_pred_proba, number_of_positives)
    plt.savefig(os.path.join(images_path, '_'.join([str(name), 'PPRAUCcurve.pdf'])))
    if log_with_mlflow:
        mlflow.log_artifact(os.path.join(images_path, '_'.join([str(name), 'PPRAUCcurve.pdf'])))



    # catched@1000
    positives_at = get_true_positives_at(y_true, y_pred_proba, number_of_positives)
    if log_with_mlflow:
        mlflow.log_metric('catched_' + str(number_of_positives), positives_at)

    performance_dict['catched_' + str(number_of_positives)] = positives_at

    return performance_dict



import numpy as np
import pandas as pd
import networkx as nx
import dateparser
import math
import pickle
import glob
from pathlib import Path
from multiprocessing import Pool
from tqdm import tqdm
from functools import partial
import os
import logging
logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)

# ~~Parameters~~ #
lambdas = {'ST': 0.03, 'MT': 0.004, 'LT': 0.0001}
data_path = '/Users/Raf/Dropbox/DOC/data/FUCC/'
files = glob.glob(os.path.join(data_path,'pagerank/pagerank_input/data_pagerank_*.csv'))
output_folder = os.path.join(data_path, 'pagerank/pagerank_output/')


def get_pagerank_suspicion_scores(data, 
                                  lambd, 
                                  alpha=0.85, 
                                  n_jobs=5,
                                  personalization_nodes=None,
                                  weighted = True,
                                  t = None):
    """[calculcate pagerank suspicion scores for every rows in data]

    Arguments:
        data {[pandas DataFrame]} -- [description]

    Keyword Arguments:
        t {str} -- [description] (default: {'ST'})
        lambd {float} -- [description] (default: {0.03})
        alpha {float} -- [description] (default: {0.85})
        n_jobs {int} -- [description] (default: {5})
        personalization_nodes {list} -- a list of nodes that will be used in the personalization vector. If None is supplied, all fraudulent nodes
        will be used in the personalization vector. 

    Returns:
        [dict] -- [description]
        [networkx Graph] -- [description]
    """

    logging.info("Building network")


    #weights_data = calculate_edge_weights(data, t=t, lambd=lambd)
    suspicion_scores = dict()
    weights_data_tx_id_key = dict()

    if weighted:
        logging.info('Calculating edge weights for {}'.format(t))
        weights_data = calculate_edge_weights(data, t=t, lambd=lambd)

        weighted_edgelist = list(zip(data.CARD_PAN_ID, data.TX_ID,
                                    [weights_data.get(key) for key in data.index])) + \
                            list((zip(data.TERM_MIDUID, data.TX_ID,
                                    [weights_data.get(key) for key in data.index])))

        
        # determine starting vector
        # starting_vector = ... Has to be time-weighted (See p.155) - update: In the APATE paper this vector e_0 is a random vector. The personalization vector has to be time-weighted, see below.
        # I will not specifiy a non-default e_0 vector (which is referred to as nstart in networkx)

        # personalization
        # Make a new weights dictionary with TX_ID as key
        logging.info('Personalization')
        weights_data_tx_id_key = dict(zip(data.TX_ID, weights_data.values()))

    else:
        # Weighted edgelist with all 1
        weighted_edgelist = list(zip(data.CARD_PAN_ID, data.TX_ID,
                            [1 for key in data.index])) + \
                    list((zip(data.TERM_MIDUID, data.TX_ID,
                            [1 for key in data.index])))

        logging.info('Personalization')
        weights_data_tx_id_key = dict(zip(data.TX_ID, [1 for entry in data.TX_ID]))
    
    logging.info('Building graph {}'.format(t))
    G = nx.Graph()
    G.add_weighted_edges_from(weighted_edgelist)

    if personalization_nodes:
        subset = data.loc[personalization_nodes]
        personalization = \
                        {dc['TX_ID']: weights_data_tx_id_key.get(dc['TX_ID']) for dc in subset[subset.TX_FRAUD == True].to_dict(orient='records')}
    else:
        personalization = \
                        {dc['TX_ID']: weights_data_tx_id_key.get(dc['TX_ID']) for dc in data[data.TX_FRAUD == True].to_dict(orient='records')}
    print(G.number_of_nodes())
    print(len(personalization))
    # Note: personalization dictionary is normalized in networkx pagerank implementation

    # run pagerank
    logging.info('Pagerank started')
    suspicion_scores = nx.pagerank(G, alpha=alpha, personalization=personalization, weight='weight')
    logging.info('Pagerank finished')
    
    # as the postprocessing step requires network information, we return the graph G as well.
    return suspicion_scores, G


def calculate_edge_weights(data, t, lambd):
    # determine weights for ST,MT,LT
    logging.info("Calculating some weights")

    current_time = dateparser.parse(data.tail(1).TX_DATETIME.dt.date.values[0].strftime("%d/%m/%Y") + " 23:59:59", settings={'DATE_ORDER': 'DMY'})

    weights_data = dict()
    if t == 'ST':
        logging.info('ok, ST')
        weights_data = (((current_time - data.TX_DATETIME).dt.total_seconds()) / 60).to_dict()
    elif t == 'MT':
        logging.info('ok, MT')
        weights_data = (((current_time - data.TX_DATETIME).dt.total_seconds()) / (60)).to_dict()
    elif t == 'LT':
        logging.info('ok, LT')
        weights_data = (((current_time - data.TX_DATETIME).dt.total_seconds()) / (60)).to_dict()
    else:
        logging.error('The time granularity specified was not recognized.')

    weights_data = {k: math.e ** (-lambd * h) for k, h in weights_data.items()}

    return weights_data


def inductive_step(edges, historical_edges, suspicion_scores, G, t):
    """
    inductive_step is a function which will inductively generate suspicion scores for every edge in edges based on the 
    suspicion_scores of historical observations. 

    param:edges = pandas dataframe containing an edgelist of edges for which you want to calculate a suspicion score
    param:historical_edges = pandas dataframe containing an edgelist of edges observed before the edges in param:edges
    param:suspicion_scores = suspicion_scores of the edges in param:historical_edges
    param:G = networkx graph based on historical_edges
    param:t = time granularity ('ST', 'LT', 'MT')
    """
    logging.info('post processing started')

    # create some new columns to store the suspicion scores
    edges = edges.assign(
        **{'SC_TX_'+t: 0, 'SC_MC_'+t: 0, 'SC_CC_'+t: 0})

    # get suspicion scores
    for i, edge in edges.iterrows():
    
        client = edge.iloc[0]
        merchant = edge.iloc[1]

        indices_client = historical_edges.iloc[:, 0].isin([edge.iloc[0]])
        indices_merchant = historical_edges.iloc[:, 1].isin([edge.iloc[1]])

        client_subset = historical_edges.loc[indices_client]
        merchant_subset = historical_edges.loc[indices_merchant]

        idx1 = client_subset.index
        idx2 = merchant_subset.index
        common_indices = idx1.intersection(idx2)

        if np.any(common_indices):
            
            most_recent_transaction_id = common_indices[-1]
            edges.loc[i, ['SC_TX_' + t]] = \
                suspicion_scores[most_recent_transaction_id]
            edges.loc[i, ['SC_MC_' + t]] = suspicion_scores[merchant]
            edges.loc[i, ['SC_CC_' + t]] = suspicion_scores[client]


        else:
            if np.any(idx1):
                # we know the client exists and had transactions before --> hence get the score
                edges.loc[i, ['SC_CC_' + t]] = suspicion_scores[edge.iloc[0]]
                CCH_score = suspicion_scores[edge.iloc[0]]
                egonet_weights_client = G.edges(edge.iloc[0], data='weight')
                sum_of_egonet_weights_client = np.sum(list(zip(*egonet_weights_client))[2])
                term_client = (1/(sum_of_egonet_weights_client + 1)) * CCH_score
            else:
                # client_score = 0
                edges.loc[i, ['SC_CC_' + t]] = 0
                # term van client valt ook weg
                term_client = 0
            if np.any(idx2):
                # we know the merchant exists and had transactions before --> hence get the score
                edges.loc[i, ['SC_MC_' + t]] = suspicion_scores[edge.iloc[1]]
                MC_score = suspicion_scores[edge.iloc[1]]
                egonet_weights_merchant = G.edges(edge.iloc[1], data='weight')
                sum_of_egonet_weights_merchant = np.sum(list(zip(*egonet_weights_merchant))[2])
                term_merchant = (1/(sum_of_egonet_weights_merchant + 1)) * MC_score
            else:
                # merchant_score = 0
                edges.loc[i, ['SC_MC_' + t]] = 0
                # term van merchant valt weg
                term_merchant = 0
            
            edges.loc[i, ['SC_TX_'+t]] = term_client + term_merchant

    return edges

def postprocessing_historical_edges(historical_edges, suspicion_scores, t):
    """
    The purpose of this method is to gather all suspicion scores of the historical data (non-inductive) into a 
    Pandas Dataframe and save to disk. 

    Args:
        historical_edges ([Pandas DataFrame]): [description]
        suspicion_scores ([type]): [description]
        G ([Networkx Graph]): [description]
        t ([String]): lambda (ST, MT, LT)
    """
    # create some new columns to store the suspicion scores
    historical_edges = historical_edges.assign(
        **{'SC_TX_'+t: 0, 'SC_MC_'+t: 0, 'SC_CC_'+t: 0})

    # get suspicion scores
    for i, edge in historical_edges.iterrows():

        client = edge.iloc[0]
        merchant = edge.iloc[1]
        transaction = i

        historical_edges.loc[i, ['SC_TX_' + t]] = \
                suspicion_scores[transaction]
        historical_edges.loc[i, ['SC_MC_' + t]] = suspicion_scores[merchant]
        historical_edges.loc[i, ['SC_CC_' + t]] = suspicion_scores[client]
    
    return historical_edges

def splitDataFrameIntoSmaller(df, chunkSize = 10000): 
    listOfDf = list()
    numberChunks = len(df) // chunkSize + 1
    for i in range(numberChunks):
        listOfDf.append(df[i*chunkSize:(i+1)*chunkSize])
    return listOfDf


def _pagerank_processing(file):
    
    df = pd.read_csv(file, index_col=0, parse_dates=[7])
    print(df.head())
    historical_data = df.loc[df.TX_DATETIME.dt.date < df.tail(1).TX_DATETIME.dt.date.values[0]]
    data = df.loc[df.TX_DATETIME.dt.date == df.tail(1).TX_DATETIME.dt.date.values[0]]

    for t, lambd in lambdas.items():
        suspicion_scores, G = get_pagerank_suspicion_scores(
                                  historical_data,
                                  t=t,
                                  lambd=lambd,
                                  alpha=0.85,
                                  n_jobs=1)
        

        f = open(output_folder + Path(file).stem + 'pagerank_suspicion_scores_' + str(t) + '.pkl', "wb")
        pickle.dump(suspicion_scores,f)
        f.close()

        to_save = pd.DataFrame.from_dict(suspicion_scores, orient='index')
        name = output_folder + Path(file).stem + 'pagerank_suspicion_scores_' + str(t) + '.csv'
        to_save.to_csv(name)

        nx.write_gpickle(G, path = output_folder + Path(file).stem + 'graph_'+str(t)+'.pkl')

def _pagerank_postprocessing(file):

    # read the data
    df = pd.read_csv(file, index_col=0, parse_dates=[7])
    
    # split in historical data and current day data
    historical_data = df.loc[df.TX_DATETIME.dt.date < df.tail(1).TX_DATETIME.dt.date.values[0]]
    data = df.loc[df.TX_DATETIME.dt.date == df.tail(1).TX_DATETIME.dt.date.values[0]]

    # We only need these columns from data
    data = data.set_index('TX_ID')
    data = data.loc[:, ['CARD_PAN_ID', 'TERM_MIDUID']]

    # We only need these columns from historical data
    historical_data = historical_data.set_index('TX_ID')
    historical_data = historical_data.loc[:, ['CARD_PAN_ID', 'TERM_MIDUID']]

    for t,lamb in lambdas.items():
        # get the suspicion scores
        # get the graph
        filename_graph = output_folder + Path(file).stem + 'graph_' + str(t) + '.pkl'
        filename_scores = output_folder + Path(file).stem + 'pagerank_suspicion_scores_' + str(t) + '.pkl'
        
        G = nx.read_gpickle(filename_graph)
        with open(filename_scores, 'rb') as handle:
            suspicion_scores = pickle.load(handle)
        
        # split df in smaller chunks
        chunks = splitDataFrameIntoSmaller(data, chunkSize=10000)
        partial_inductive_step = partial(inductive_step, historical_edges=historical_data, suspicion_scores=suspicion_scores, G=G, t=t)

        # multiprocessing
        with Pool(8) as p:
            r = list(tqdm(p.imap(partial_inductive_step, chunks), total=len(chunks)))

        df = pd.concat(r)
        df.to_csv(output_folder + Path(file).stem + '_output_'+str(t)+'.csv')

        #Split historical dataset into smaller chunks
        chunks = splitDataFrameIntoSmaller(historical_data, chunkSize=50000)
        partial_postprocessing_historical_edges = partial(postprocessing_historical_edges, suspicion_scores=suspicion_scores, t=t)
        
        # multiprocessing
        with Pool(8) as p:
            r = list(tqdm(p.imap(partial_postprocessing_historical_edges, chunks), total=len(chunks)))

        df = pd.concat(r)
        df.to_csv(output_folder + Path(file).stem + '_output_historical_edges_'+str(t)+'.csv')

        p.close()



if __name__ == '__main__':
    
    #with Pool(8) as p:
    #   r = list(tqdm(p.imap(_pagerank_processing, files), total=len(files)))


    for fil in tqdm(files, total=len(files)):
        _pagerank_postprocessing(fil)



from multiprocessing import Pool
import pandas as pd
from functools import partial
import numpy as np
from tqdm import tqdm


def inductive_pooling(df, embeddings, G, workers, gamma=1000, dict_node=None, average_embedding=True):
    


    if average_embedding:
        avg_emb = embeddings.mean().values
    else:
        avg_emb = None
    
    #if __name__ == '__main__':
    with Pool(workers) as p:
        r = p.map(partial(inductive_pooling_chunk, embeddings=embeddings, G=G, average_embedding=avg_emb), np.array_split(df, workers))

    return r
	
def inductive_pooling_chunk(df, embeddings, G, gamma=1000, average_embedding=None):
	
    #Create a container for the new embeddings
    new_embeddings = dict()

    for transaction, transaction_row in tqdm(df.iterrows(), total=df.shape[0]):

        cardholder = transaction_row.CARD_PAN_ID
        merchant = transaction_row.TERM_MIDUID


        mutual = False

        if G.has_node(cardholder) & G.has_node(merchant):
            mutual_neighbors = list(set(G.neighbors(cardholder)).intersection(set(G.neighbors(merchant))))
            mutual_neighbors.sort()
            if (len(mutual_neighbors) > 0): 
                mutual = True
                # Use dataframe with TX_ID on index (to speed up retrieval of transaction rows)
                embeddings_mutual_neighbors = embeddings.loc[mutual_neighbors]
                # most recent transaction
                most_recent_embedding_mutual_neighbor = embeddings_mutual_neighbors.iloc[-1]

                new_embeddings[transaction] = most_recent_embedding_mutual_neighbor
                
                        
        if G.has_node(cardholder) & (not mutual):

            cardholder_neighbors = list(G.neighbors(cardholder))
            
            pooled_embedding = get_pooled_embedding(cardholder_neighbors, embeddings, gamma)
            
            new_embeddings[transaction] = pooled_embedding
            
        elif G.has_node(merchant) & (not mutual):

            merchant_neighbors = list(G.neighbors(merchant))

            pooled_embedding = get_pooled_embedding(merchant_neighbors, embeddings, gamma)

            new_embeddings[transaction] = pooled_embedding
            
            
        elif (not mutual):
            new_embeddings[transaction] = average_embedding
                    

    return new_embeddings
								
def get_pooled_embedding(neighbors, embeddings, gamma):
	
	embeddings_to_pool = embeddings.loc[neighbors]
	most_recent_embeddings_to_pool = embeddings_to_pool.iloc[-min(gamma, embeddings_to_pool.shape[0]):]
	
	pooled_embedding = pd.DataFrame(most_recent_embeddings_to_pool.mean()).transpose().values[0]
	
	return pooled_embedding
	




import networkx as nx
import logging
from multiprocessing import Pool
import os
import pickle
import pandas as pd
from tqdm import tqdm

# Create a custom logger
logger = logging.getLogger("utils")


def invert_dict(d):
    return dict([(v, k) for k, v in d.items()])

def export_network(network, output_filepath):
    logger.info("Converting node labels to integers")
    G = nx.convert_node_labels_to_integers(network, first_label=0, ordering='default', label_attribute='label')
    logger.info("Exporting edgelist")
    logger.info(output_filepath)
    nx.write_edgelist(G, output_filepath, comments='#', delimiter=' ', data=False, encoding='utf-8')

    node_dict = dict(G.nodes(data='label'))
    dict_node = invert_dict(node_dict)

    return dict_node

def get_filename(filename_elements = []):
    filename = '_'.join([str(x) for x in filename_elements])
    return filename

def export_suspicion_scores(suspicion_scores, G, filename_elements = [], output_filepath=''):
    
    filename = get_filename(filename_elements=filename_elements)
    
    # Export raw suspicion scores
    f = open(os.path.join(output_filepath, filename + '_pagerank_suspicion_scores.pkl'), "wb")
    pickle.dump(suspicion_scores,f)
    f.close()

    # Export suspicion scores in Pandas DataFrame CSV
    to_save = pd.DataFrame.from_dict(suspicion_scores, orient='index')
    name = os.path.join(output_filepath, filename + '_pagerank_suspicion_scores.csv')
    to_save.to_csv(name)

    # Export Graph
    nx.write_gpickle(G, path = os.path.join(output_filepath, filename + '_graph.pkl'))



def import_suspicion_scores(filename_elements = [], input_filepath=''):
    
    filename = get_filename(filename_elements=filename_elements)
    
    # get the suspicion scores
    # get the graph
    filename_graph = os.path.join(input_filepath, filename + '_graph.pkl')
    filename_scores = os.path.join(input_filepath, filename + '_pagerank_suspicion_scores.pkl')

    G = nx.read_gpickle(filename_graph)
    with open(filename_scores, 'rb') as handle:
        suspicion_scores = pickle.load(handle)
        
    return suspicion_scores, G

def multiprocessing(function, chunks, workers=8):
        
    # multiprocessing
    with Pool(workers) as p:
        r = list(tqdm(p.imap(function, chunks), total=len(chunks)))

    p.close()
    concatenated_result = pd.concat(r)
    return concatenated_result



#from sklearn.model_selection._split import _BaseKFold
from sklearn.utils import indexable
from sklearn.utils.validation import _deprecate_positional_args, _num_samples
import datetime 
import numpy as np
__all__ = ['RollingWindowSplit']

class RollingWindowSplit():
    
    @_deprecate_positional_args
    def __init__(self, train_size=None, test_size=None, start_date=None):
        
                  
        self.train_size = train_size
        self.test_size = test_size
        self.start_date = start_date

        # If these are specified, check they belong to the right type. 
        #if (train_size is not None) & (not isinstance(train_size, datetime.timedelta)):
        #    raise TypeError('train_size should be datetime.timedelta')
        #if (test_size is not None ) & (not isinstance(test_size, datetime.timedelta)):
        #    raise TypeError('test_size should be datetime.timedelta')


    def split(self, X, y=None, timestamps=None):
        """Generate indices to split data into training and test set.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like of shape (n_samples,)
            Always ignored, exists for compatibility.
        timestamps : array-like of shape (n_samples,)
            containing timestamp associated with every sample.
        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """

        X, y, timestamps = indexable(X, y, timestamps)
        n_samples = _num_samples(X)


        if (self.train_size is not None) & (self.test_size is not None):
            
            indices = np.arange(n_samples)
            if self.start_date is not None:
                if self.start_date < timestamps[0] + self.train_size:
                    raise ValueError(
                    "The provided start_date is too close to the start of the datasets to allow for sufficient training data.")
                
                start_time = self.start_date
            else:
                start_time = timestamps[0] + self.train_size
            
            end_time = timestamps[-1]
            
            while start_time + self.test_size <= end_time:
                yield (indices[(timestamps > start_time - self.train_size) & (timestamps <= start_time)], 
                    indices[(timestamps > start_time) & (timestamps <= (start_time + self.test_size))])
                # Update start_time
                start_time += self.test_size

    def number_of_splits(self, timestamps=None):
        """Calculate the number of splits possible within the given dataset

        Args:
            timestamps ([type], optional): [description]. Defaults to None.
        """

        return np.floor((timestamps[-1] - self.start_date) / self.test_size)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fucc", 
    version="0.0.6",
    author="Rafael Van Belle",
    author_email="rafael@gmail.com",
    description="Fraudulent Use of Credit Card",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rafaelvanbelle/fucc_package",
    packages=setuptools.find_packages(),
    install_requires=[
          'scikit-learn', 
          'pandas', 
          'tqdm', 
          'scikit-plot', 
          'mlflow', 
          'matplotlib', 
          'networkx', 
          'dateparser'
      ],
    classifiers=[
        "Programming Language :: Python :: 3", 
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)